import shutil
import os
from pathlib import Path
from datetime import datetime


class DaisysMcpError(Exception):
    pass


def throw_mcp_error(message: str):
    raise DaisysMcpError(message)


def is_installed(lib_name: str) -> bool:
    lib = shutil.which(lib_name)
    if lib is None:
        return False
    return True


def is_file_writeable(path: Path) -> bool:
    if path.exists():
        return os.access(path, os.W_OK)
    parent_dir = path.parent
    return os.access(parent_dir, os.W_OK) and parent_dir.exists()


def make_output_file(text: str, output_path: Path, extension: str = "wav") -> Path:
    text = text.replace(" ", "_")
    output_file_name = (
        f"{text[:5]}_{datetime.now().strftime('%Y%m%d_%H%M%S')}.{extension}"
    )
    return output_path / output_file_name


def make_output_path(
    output_directory: str | None, base_path: str | None = None
) -> Path:
    output_path = None
    if output_directory is None:
        output_path = Path.home() / "Desktop"
    elif not os.path.isabs(output_directory) and base_path:
        output_path = Path(os.path.expanduser(base_path)) / Path(output_directory)
    else:
        output_path = Path(os.path.expanduser(output_directory))

    output_path.mkdir(parents=True, exist_ok=True)

    if not is_file_writeable(output_path):
        throw_mcp_error(f"Directory ({output_path}) is not writeable")
    return output_path



from pydantic import BaseModel  # type: ignore
from typing import List, Optional
import enum


class Status:
    WAITING = "waiting"
    STARTED = "started"
    READY = "ready"
    ERROR = "error"
    TIMEOUT = "timeout"


class McpVoice(BaseModel):
    voice_id: str
    name: str
    gender: str
    model: str
    description: str | None


class McpModel(BaseModel):
    name: str
    displayname: str
    flags: Optional[List]
    languages: List[str]
    genders: List[str]
    styles: List[List[str]]
    prosody_types: List[str]


class VoiceGender(str, enum.Enum):
    """Represents the gender of a voice.

    Note: upper case in Python, lower case in JSON.

    Values:
      MALE, FEMALE, NONBINARY
    """

    MALE = "male"
    FEMALE = "female"
    NONBINARY = "nonbinary"



import os
import io
from typing import Optional


from daisys import DaisysAPI  # type: ignore
from daisys.v1.speak import SimpleProsody, DaisysTakeGenerateError  # type: ignore

from daisys_mcp.utils import throw_mcp_error

disable_audio_playback = os.getenv("DISABLE_AUDIO_PLAYBACK", "false").lower() == "true"

email = os.environ.get("DAISYS_EMAIL")
password = os.environ.get("DAISYS_PASSWORD")


def text_to_speech_http(text: str, voice_id: Optional[str] = None):
    """
    Generate and play audio from text using DaisysAPI's HTTP protocol with sounddevice.
    """
    if not email or not password:
        throw_mcp_error(
            "DAISYS_EMAIL and DAISYS_PASSWORD environment variables must be set."
        )

    if text in ["None", "", None]:
        throw_mcp_error("Text for TTS cannot be empty.")

    with DaisysAPI("speak", email=email, password=password) as speak:
        try:
            take = speak.generate_take(
                voice_id=voice_id,
                text=text,
                prosody=SimpleProsody(pace=0, pitch=0, expression=5),
            )
        except DaisysTakeGenerateError as e:
            raise RuntimeError(f"Error generating take: {str(e)}")

        audio_mp3 = speak.get_take_audio(take.take_id, format="mp3")

        if not disable_audio_playback:
            try:
                import sounddevice as sd  # type: ignore
                import soundfile as sf  # type: ignore
            except ModuleNotFoundError:
                message = (
                    "`uv pip install sounddevice soundfile` to enable audio playback."
                )
                raise ValueError(message)
            sd.play(*sf.read(io.BytesIO(audio_mp3)))
            sd.wait()

        return audio_mp3



import os
from daisys import DaisysAPI  # type: ignore

# from daisys.v1.speak.models import ProsodyFeaturesUnion, ProsodyType
from mcp.server.fastmcp import FastMCP  # type: ignore
from mcp.types import TextContent
from typing import Literal

from daisys_mcp.model import McpVoice, McpModel, VoiceGender
from daisys_mcp.websocket_tts import text_to_speech_websocket
from daisys_mcp.http_tts import text_to_speech_http
from daisys_mcp.utils import throw_mcp_error, make_output_file, make_output_path

from dotenv import load_dotenv  # type: ignore

load_dotenv()

# Initialize FastMCP server
mcp = FastMCP("Daisys-mcp-server")
email = os.environ.get("DAISYS_EMAIL")
password = os.environ.get("DAISYS_PASSWORD")

if not email or not password:
    throw_mcp_error("DAISYS_EMAIL, DAISYS_PASSWORD environment variable is required")

storage_path = os.environ.get("DAISYS_BASE_STORAGE_PATH")


@mcp.tool(
    "text_to_speech",
    description=(
        """
        Convert text to speech with a given voice and play the audio buffer.
        Voice_id can be provided. If not provided, the latest voice will be used.

        ⚠️ TOKEN WARNING: This tool makes an API call to Daisys API which may incur costs. 

        Args:
            text (str): The text to convert to speech.
            voice_id (str, optional): The voice_id of the voice to use. If no voice specified use latest created voice.
            audio_format (str, optional): Can be either "wav" or "mp3". Defaults to "wav" always use "wav" unless mp3 specified.
            output_dir (str, optional): Directory where files should be saved. Defaults to $HOME/Desktop if not provided.
            streaming (bool, optional): Whether to use streaming or not. Set to True unless specifically asked to not stream. (streaming makes use of the websocket protocol which send and play audio in chunks)
            Defaults don't store if not provided.

        Returns:
            Text content with the path to the output file and name of the voice used.
        """
    ),
)
# Disabled optional typing since its not yet supported by cursor's mcp client
def text_to_speech(
    text: str,
    voice_id: str = None,  # type: ignore
    audio_format: str = "wav",
    output_dir: str = None,  # type: ignore
    streaming: bool = True,
):
    if text in ["None", "", None]:
        throw_mcp_error("Text for TTS cannot be empty.")

    # LLM sometimes send null as a string
    if isinstance(voice_id, str) and voice_id.lower() in ["null", "undefined"]:
        voice_id = None  # type: ignore

    with DaisysAPI("speak", email=email, password=password) as speak:  # type: ignore
        if not voice_id:
            try:
                voice_id = speak.get_voices()[-1].voice_id
            except IndexError:
                throw_mcp_error("No voices available. Try to generate a voice first.")

    try:
        # this can create only a wav file but has fast inference
        if audio_format == "wav" and streaming:
            audiobuffer = text_to_speech_websocket(text, voice_id)
        else:
            audiobuffer = text_to_speech_http(text, voice_id)
    except Exception:
        throw_mcp_error("Error generating audio")

    if not storage_path:
        return TextContent(
            type="text",
            text=f"Success. Voice used: {voice_id}",
        )
    # Create the output file
    output_path = make_output_path(output_dir, storage_path)
    output_file_name = make_output_file(text, output_path, audio_format)

    output_path.parent.mkdir(parents=True, exist_ok=True)
    with open(output_path / output_file_name, "wb") as f:
        f.write(audiobuffer)  # type: ignore

    return TextContent(
        type="text",
        text=f"Success. File saved as: {output_path / output_file_name}. Voice used: {voice_id}",
    )


@mcp.tool(
    "get_voices",
    description="Get available voices can be filtered by model and gender, and sorted by name or timestamp in ascending or descending order.",
)
# Disabled optional typing since its not yet supported by cursor's mcp client
def get_voices(
    model: str = None,
    gender: str = None,
    sort_by: Literal["description", "name"] = "name",
    sort_direction: Literal["asc", "desc"] = "asc",
):
    with DaisysAPI("speak", email=email, password=password) as speak:
        filtered_voices = [
            voice
            for voice in speak.get_voices()
            if (model is None or voice.model == model)
            and (gender is None or voice.gender == gender)
        ]
        voice_list = [
            McpVoice(
                voice_id=voice.voice_id,
                name=voice.name,
                gender=voice.gender,
                model=voice.model,
                description=voice.description,
            )
            for voice in filtered_voices
        ]
        if sort_direction == "asc":
            voice_list.sort(key=lambda x: getattr(x, sort_by).lower())

        else:
            voice_list.sort(key=lambda x: getattr(x, sort_by).lower(), reverse=True)

        return voice_list


@mcp.tool(
    "get_models",
    description=(
        """
        Get all available models from Daisys API.

        Args:
            language (str, optional): needs to be "de" for german, "en" for english and "nl" for dutch. Defaults to None.
            sort_by (str, optional): can be "name" or "displayname". Defaults to "displayname".
            sort_direction (str, optional): can be "asc" or "desc". Defaults to "asc".

        Returns:
            model_list: An object containing details of all models

        """
    ),
)

# Disabled optional typing since its not yet supported by cursor's mcp client
def get_models(
    language: str = None,
    sort_by: Literal["name", "displayname"] = "displayname",
    sort_direction: Literal["asc", "desc"] = "asc",
):
    # make sure to only use the first 2 letters of the language
    if language:
        language = language.lower()[:1]

    with DaisysAPI("speak", email=email, password=password) as speak:
        filtered_models = [
            model
            for model in speak.get_models()
            if language is None
            or any(lang.startswith(language) for lang in model.languages)
        ]
        model_list = [
            McpModel(
                name=model.name,
                displayname=model.displayname,
                flags=model.flags,
                languages=model.languages,
                genders=model.genders,
                styles=model.styles,
                prosody_types=model.prosody_types,
            )
            for model in filtered_models
        ]

        if sort_direction == "asc":
            model_list.sort(key=lambda x: getattr(x, sort_by).lower())

        else:
            model_list.sort(key=lambda x: getattr(x, sort_by).lower(), reverse=True)
        return model_list


@mcp.tool(
    "create_voice",
    description=(
        """
        Convert text to speech with a given voice and play the audio buffer.
        Voice_id can be provided. If not provided, the latest voice will be used.

        before calling also call get models so it always inputs a valid model

        ⚠️ TOKEN WARNING: This tool makes an API call to Daisys API which may incur costs. 

        Args:
            name (str, optional): The name of the voice to create. Defaults to "Daisy".
            gender (str, optional): The gender of the voice can be "male" or "female". Defaults to "female".
            model (str, optional): The model of the voice. Defaults to "english-v3.0".
            pitch (int, optional): Adjusts the pitch level; -10 for very low pitch, 10 for very high pitch.
            pace (int, optional): Controls the speech rate; -10 for very slow, 10 for very fast.
            expression (int, optional): Modulates expressiveness; -10 for monotone, 10 for highly expressive.

        Returns:
            McpVoice: An object containing details of the created voice, including voice_id, name, gender, model, and description.

        """
    ),
)
def create_voice(
    name: str = "Daisy",
    gender: str = "female",
    model: str = "english-v3.0",
    pitch: int = 0,
    pace: int = 0,
    expression: int = 0,
):
    # Using SimpleProsody
    prosody_params = {
        "pitch": pitch,
        "pace": pace,
        "expression": expression,
    }
    # Validate that all prosody parameters are within the range -10 to 10
    for param_name, value in prosody_params.items():
        if not -10 <= value <= 10:
            throw_mcp_error(f"{param_name} must be between -10 and 10, got {value}.")

    if gender not in VoiceGender:
        throw_mcp_error(
            f"Invalid gender: {gender}. Must be one of {list(VoiceGender)}."
        )

    with DaisysAPI("speak", email=email, password=password) as speak:
        voice = speak.generate_voice(
            name=name, gender=gender, model=model, default_prosody=prosody_params
        )
    return McpVoice(
        voice_id=voice.voice_id,
        name=voice.name,
        gender=voice.gender,
        model=voice.model,
        description=voice.description,
    )


@mcp.tool(
    "remove_voice",
    description="Delete a voice.",
)
def remove_voice(
    voice_id: str,
):
    with DaisysAPI("speak", email=email, password=password) as speak:
        speak.delete_voice(voice_id)

    return TextContent(
        type="text",
        text=f"Success. voice {voice_id} deleted.",
    )


def main():
    print("Starting Daisys-mcp server.")
    mcp.run(transport="stdio")


if __name__ == "__main__":
    main()



import os
import time
import numpy as np  # type: ignore
import sounddevice as sd  # type: ignore
import wave
from typing import Optional

from daisys import DaisysAPI  # type: ignore
from daisys.v1.speak import (  # type: ignore
    DaisysWebsocketGenerateError,
    Status,
    StreamOptions,
    StreamMode,
)

from daisys_mcp.utils import throw_mcp_error
import io

from dotenv import load_dotenv  # type: ignore

load_dotenv()

disable_audio_playback = os.getenv("DISABLE_AUDIO_PLAYBACK", "false").lower() == "true"
email = os.environ.get("DAISYS_EMAIL")
password = os.environ.get("DAISYS_PASSWORD")
storage_path = os.environ.get("DAISYS_BASE_STORAGE_PATH")


def text_to_speech_websocket(text: str, voice_id: Optional[str] = None):
    """
    Generate and return WAV audio from text using DaisysAPI's WebSocket protocol.
    """

    if not email or not password:
        raise throw_mcp_error("DAISYS_EMAIL and DAISYS_PASSWORD must be set.")

    wav_buffer = io.BytesIO()
    audio_chunks = []
    stream = None

    if not disable_audio_playback:
        stream = sd.OutputStream(
            samplerate=22050,  # or check from the actual stream
            channels=1,
            dtype="int16",
        )
        stream.start()

    with DaisysAPI("speak", email=email, password=password) as speak:
        with speak.websocket(voice_id=voice_id) as ws:
            done = False
            ready = False
            generated_take = None
            t0 = time.time()

            def audio_cb(request_id, take_id, part_id, chunk_id, audio):
                nonlocal done

                if audio:
                    audio_np = np.frombuffer(audio, dtype=np.int16)
                    if not disable_audio_playback:
                        stream.write(audio_np) if stream else None
                    audio_chunks.append(audio)

                else:
                    if chunk_id in [0, None]:
                        done = True

            def status_cb(request_id, take):
                nonlocal ready, generated_take
                generated_take = take
                if take.status == Status.READY:
                    ready = True

            ws.generate_take(
                voice_id=voice_id,
                text=text,
                status_callback=status_cb,
                audio_callback=audio_cb,
                stream_options=StreamOptions(mode=StreamMode.CHUNKS),
            )

            while not (ready and done) and (time.time() - t0) < 60:
                try:
                    ws.update(timeout=5)
                except DaisysWebsocketGenerateError as e:
                    throw_mcp_error(e)
                    break

    # Combine audio chunks and write to wav buffer
    combined_audio = b"".join(audio_chunks)
    with wave.open(wav_buffer, "wb") as wav_file:
        wav_file.setnchannels(1)
        wav_file.setsampwidth(2)
        wav_file.setframerate(22050)
        wav_file.writeframes(combined_audio)

    if not disable_audio_playback:
        stream.stop() if stream else None
        stream.close() if stream else None

    return wav_buffer.getvalue()
